;; Generated package description from company-statistics.el
(define-package "company-statistics" "0.1.1" "Sort candidates using completion history" '((emacs "24.3") (company "0.8.5")) :url "https://github.com/company-mode/company-statistics" :keywords '("abbrev" "convenience" "matching"))
